use std::{error::Error, sync::Arc};

use vulkano::{
    buffer::{BufferContents, IndexBuffer, Subbuffer},
    command_buffer::{AutoCommandBufferBuilder, PrimaryAutoCommandBuffer},
    descriptor_set::allocator::StandardDescriptorSetAllocator,
    pipeline::GraphicsPipeline,
};

use super::AsDescriptorSet;

pub trait AsRecordable {
    fn record_bind_commands(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        descriptor_set_allocator: &Arc<StandardDescriptorSetAllocator>,
        pipeline: &Arc<GraphicsPipeline>,
        mesh: &impl AsRenderableMesh,
        instances: &impl AsRenderableMeshInstance,
        descriptor_sets: Vec<Arc<dyn AsDescriptorSet>>,
    ) -> Result<(), Box<dyn Error>>;

    fn record_draw_commands(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        mesh: &impl AsRenderableMesh,
        instances: &impl AsRenderableMeshInstance,
    ) -> Result<(), Box<dyn Error>> {
        match mesh.index_buffer() {
            Some(index_buffer) => {
                builder.bind_index_buffer(index_buffer.clone())?;
                unsafe {
                    builder.draw_indexed(
                        mesh.index_count(),
                        instances.instance_count(),
                        mesh.first_index(),
                        mesh.vertex_offset(),
                        instances.first_instance(),
                    )?;
                }
            }
            None => unsafe {
                builder.draw(
                    mesh.vertex_count(),
                    instances.instance_count(),
                    mesh.first_vertex(),
                    instances.first_instance(),
                )?;
            },
        }

        Ok(())
    }

    fn record_commands(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        descriptor_set_allocator: &Arc<StandardDescriptorSetAllocator>,
        pipeline: &Arc<GraphicsPipeline>,
        mesh: &impl AsRenderableMesh,
        instances: &impl AsRenderableMeshInstance,
        descriptor_sets: Vec<Arc<dyn AsDescriptorSet>>,
    ) -> Result<(), Box<dyn Error>> {
        Self::record_bind_commands(
            builder,
            descriptor_set_allocator,
            pipeline,
            mesh,
            instances,
            descriptor_sets,
        )?;
        Self::record_draw_commands(builder, mesh, instances)?;
        Ok(())
    }
}

pub trait AsRenderableMesh {
    type VertexBufferData: BufferContents + Clone;
    type IndexBuffer: Into<IndexBuffer> + Clone;

    fn vertex_buffer(&self) -> &Subbuffer<[Self::VertexBufferData]>;

    fn vertex_count(&self) -> u32;

    fn first_vertex(&self) -> u32 {
        0
    }

    fn vertex_offset(&self) -> i32 {
        0
    }

    fn index_buffer(&self) -> Option<&Self::IndexBuffer> {
        None
    }

    fn index_count(&self) -> u32 {
        0
    }

    fn first_index(&self) -> u32 {
        0
    }
}

pub trait AsRenderableMeshInstance {
    type InstanceBufferData: BufferContents + Clone;

    fn instance_buffer(&self) -> &Subbuffer<[Self::InstanceBufferData]>;

    fn instance_count(&self) -> u32;

    fn first_instance(&self) -> u32 {
        0
    }
}
