use std::{error::Error, path::PathBuf, sync::Arc};

use vulkano::{buffer::Subbuffer, memory::allocator::StandardMemoryAllocator};

use crate::core::render::primitives::{
    AsIndexBuffer, AsRenderableMesh, AsVertexBuffer, vertex::Vertex3D,
};

pub struct ObjMesh {
    vertex_buffer: Subbuffer<[Vertex3D]>,
    index_buffer: Subbuffer<[u32]>,
}

impl ObjMesh {
    pub fn new(
        memory_allocator: &Arc<StandardMemoryAllocator>,
        file_path: impl Into<PathBuf>,
    ) -> Result<Vec<Self>, Box<dyn Error>> {
        let (models, _) = tobj::load_obj(
            &file_path.into(),
            &tobj::LoadOptions {
                single_index: true,
                triangulate: true,
                ..Default::default()
            },
        )?;

        let meshes = models
            .into_iter()
            .map(|model| {
                let vertices = (0..model.mesh.positions.len() / 3)
                    .map(|i| Vertex3D {
                        position: [
                            model.mesh.positions[i * 3],
                            model.mesh.positions[i * 3 + 1],
                            model.mesh.positions[i * 3 + 2],
                        ],
                        uv: [model.mesh.texcoords[i * 2], model.mesh.texcoords[i * 2 + 1]],
                    })
                    .collect::<Vec<_>>();

                let indices = model.mesh.indices;

                let vertex_buffer = Vertex3D::create_vertex_buffer(memory_allocator, &vertices)
                    .expect("Failed to create vertex buffer");
                let index_buffer = u32::create_index_buffer(memory_allocator, &indices)
                    .expect("Failed to create index buffer");

                Self {
                    vertex_buffer,
                    index_buffer,
                }
            })
            .collect::<Vec<_>>();

        Ok(meshes)
    }
}

impl AsRenderableMesh for ObjMesh {
    type VertexBufferData = Vertex3D;
    type IndexBuffer = Subbuffer<[u32]>;

    fn vertex_buffer(&self) -> &Subbuffer<[Self::VertexBufferData]> {
        &self.vertex_buffer
    }

    fn index_buffer(&self) -> Option<&Self::IndexBuffer> {
        Some(&self.index_buffer)
    }

    fn vertex_count(&self) -> u32 {
        self.vertex_buffer.len() as u32
    }

    fn index_count(&self) -> u32 {
        self.index_buffer.len() as u32
    }
}
